# Copyright: Hallux team, 2023

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Final

from ..proposals.proposal_engine import ProposalEngine


class IssueDescriptor(ABC):
    def __init__(
        self,
        tool: str,
        filename: str,
        issue_line: int = 0,
        description: str = "",
        issue_type: str = "warning",
        language: str | None = None,
    ):
        self.tool: Final[str] = tool
        self.filename: Final[str] = filename
        self.issue_line: int = issue_line
        self.description: str = description
        self.message_lines: list[str] = []
        self.issue_type: str = issue_type

        if language is None:
            if filename.endswith(".py"):
                language = "python"
            elif (
                filename.endswith(".cpp")
                or filename.endswith(".hpp")
                or filename.endswith(".c")
                or filename.endswith(".h")
            ):
                language = "cpp"
            elif filename.endswith(".js") or filename.endswith(".ts"):
                language = "javascript"
            else:
                language = "programming"

        self.language: Final[str] = language

    @abstractmethod
    def list_proposals(self) -> ProposalEngine:
        # ToDo: we can provide QueryBackend or even a list[QueryBackend] in here
        # (or some kind of prioritized container)
        # This might give ability to try different Backends for different Proposals
        pass



# Copyright: Hallux team, 2023

from __future__ import annotations

import os
import subprocess
from typing import Final

import requests
from unidiff import PatchSet

from ..logger import logger
from ..proposals.diff_proposal import DiffProposal
from .filesystem import FilesystemTarget
from .github_suggestion import GithubSuggestion


# Saves Issue Fixes as Github proposals
class GitlabSuggestion(FilesystemTarget):
    def __init__(self, mr_url: str | None, verify: bool = False):
        FilesystemTarget.__init__(self)

        if "GITLAB_TOKEN" not in os.environ.keys():
            raise SystemError("GITLAB_TOKEN is not provided")

        answ = GitlabSuggestion.parse_mr_url(mr_url)
        if answ is None:
            logger.warning(f"Cannot parse gitlab MR URL: {mr_url}")
            raise SystemError(f"Cannot parse gitlab MR URL: {mr_url}")
        (base_url, project_name, mr_iid) = answ
        self.base_url: Final[str] = base_url
        self.project_name: Final[str] = project_name
        self.mr_iid: Final[int] = mr_iid
        self.verify = verify

        request_url: str = f"{base_url}/projects/{project_name}/merge_requests/{mr_iid}/changes?unidiff=true"
        headers = {"PRIVATE-TOKEN": f"{os.environ['GITLAB_TOKEN']}"}
        mr_response = requests.get(request_url, headers=headers, verify=self.verify)

        if mr_response.status_code != 200:
            logger.error(mr_response.text)
            raise SystemError(mr_response.text)

        self.mr_json = mr_response.json()
        self.base_sha: Final[str] = self.mr_json["diff_refs"]["base_sha"]
        self.head_sha: Final[str] = self.mr_json["diff_refs"]["head_sha"]
        self.start_sha: Final[str] = self.mr_json["diff_refs"]["start_sha"]

        # mapping from new_path to old_path (if file is changed)
        self.changed_files: Final[dict[str, str]] = {}
        # mapping for diffs of changed files
        self.changed_diffs: Final[dict[str, str]] = {}
        for change in self.mr_json["changes"]:
            new_path = change["new_path"]
            if new_path is not None:  # new_path is None when file is deleted
                self.changed_files[new_path] = change["old_path"]
                self.changed_diffs[new_path] = change["diff"]

        if self.mr_json["merged_at"] is not None or self.mr_json["closed_at"] is not None:
            raise SystemError(f"Merge Request {mr_iid} is either closed or merged already")

        local_git_sha: str
        try:
            git_log: str = subprocess.check_output(["git", "log", "--pretty=oneline"]).decode("utf8")
            local_git_sha = git_log.split("\n")[0].split(" ")[0]
        except Exception as e:
            raise SystemError(f"Local git commit is not available: {e}")

        if local_git_sha != self.head_sha:
            raise SystemError(
                f"Local git commit: `{local_git_sha}` "
                f"do not coinside with the head from pull-request: `{self.head_sha}` "
            )

    @staticmethod
    def parse_mr_url(mr_url: str) -> tuple[str, str, int] | None:
        """
        Tries to parse Pull-Request URL to obtain base_url, repo_name and PR_ID
        :param mr_url:
            shall a gitlab merge request URL: https://gitlab.com/hallux/hallux/-/merge_requests/2
            or a gitlab API URL: https://gitlab.com/api/v4/hallux%2Fhallux/merge_requests/2

        :return: (base_url, project_name, mr_iid) OR None
        """

        url_protocol = "https://"
        if not (mr_url.startswith(url_protocol)):
            return None

        # if mr_url includes /api/v4 - then it is an API URL
        api_prefix = "/api/v4"

        if mr_url.count(api_prefix) > 0:
            base_url = mr_url[: mr_url.find(api_prefix) + len(api_prefix)]
            project_name = mr_url[mr_url.find(api_prefix) + len(api_prefix) + 1 : mr_url.find("/merge_requests") - 2]
            mr_iid = int(mr_url[mr_url.rfind("/") + 1 :])
            return base_url, project_name, mr_iid

        mr_url = mr_url[len(url_protocol) :]

        url_items = mr_url.split("/")
        if len(url_items) == 6:
            if url_items[3] == "-" and url_items[4] == "merge_requests":
                base_url = url_protocol + url_items[0] + api_prefix
                project_name = url_items[1] + "%2F" + url_items[2]
                mr_iid = int(url_items[5])
                return base_url, project_name, mr_iid

        return None

    def apply_diff(self, diff: DiffProposal) -> bool:
        for file in self.changed_files:
            if diff.filename == file or file.endswith(diff.filename):
                FilesystemTarget.apply_diff(self, diff)
                return True
        return False

    def revert_diff(self):
        FilesystemTarget.revert_diff(self)

    @staticmethod
    def find_old_code_line(diff: str, start_line) -> int | None:
        patch_set = PatchSet(diff)
        for hunk in patch_set[0]:
            if start_line < hunk.target_start:
                return hunk.source_start + start_line - hunk.target_start

            if hunk.target_start <= start_line and start_line <= hunk.target_start + hunk.target_length:
                for line in hunk:
                    if line.target_line_no == start_line:
                        return line.source_line_no

        last_known_line = patch_set[0][-1][-1]
        return last_known_line.source_line_no + start_line - last_known_line.target_line_no

    def write_suggestion(self, proposal: DiffProposal) -> bool:
        data = {}
        data["position[position_type]"] = "text"
        data["position[base_sha]"] = self.base_sha
        data["position[head_sha]"] = self.head_sha
        data["position[start_sha]"] = self.start_sha
        data["position[new_path]"] = proposal.filename
        data["position[new_line]"] = proposal.start_line

        if self.changed_files[proposal.filename] is not None:
            data["position[old_path]"] = self.changed_files[proposal.filename]
            git_diff: str = self.changed_diffs[proposal.filename]
            old_line = self.find_old_code_line(git_diff, proposal.start_line)
            if old_line:
                data["position[old_line]"] = old_line

        body = proposal.description + "\n```suggestion"
        if proposal.end_line > proposal.start_line:
            body += f":-0+{proposal.end_line-proposal.start_line}"
        body += "\n"
        for i in range(len(proposal.proposed_lines)):
            line = proposal.proposed_lines[i]
            body += line
            body += "\n" if i < len(proposal.proposed_lines) - 1 else ""
        body = body + "\n```"
        data["body"] = body

        request_url: str = f"{self.base_url}/projects/{self.project_name}/merge_requests/{self.mr_iid}/discussions"
        headers = {"PRIVATE-TOKEN": f"{os.environ['GITLAB_TOKEN']}"}  # , "Content-Type": "text"
        mr_response = requests.post(request_url, headers=headers, data=data, verify=self.verify)

        logger.debug(mr_response.status_code)
        logger.debug(mr_response.text)
        if mr_response.status_code != 200 and mr_response.status_code != 201:
            logger.error(mr_response.text)
            return False
        return True

    def commit_diff(self) -> bool:
        success: bool = False
        try:
            compacted = GithubSuggestion.compact_proposal(self.existing_proposal)
            success = self.write_suggestion(compacted)
        finally:
            # clean local code
            FilesystemTarget.revert_diff(self)

        return success

    def requires_refresh(self) -> bool:
        return False



# Copyright: Hallux team, 2023

from __future__ import annotations

import copy
import difflib
from typing import Final

from ..backends.query_backend import QueryBackend
from ..issues.issue import IssueDescriptor
from ..targets.diff import DiffTarget
from .diff_proposal import DiffProposal


# Struct to keep all relevant info about found code fix in one place
class SimpleProposal(DiffProposal):
    def __init__(
        self,
        issue: IssueDescriptor,
        # radius or tuple with [start_line, end_line]
        radius_or_range: int | tuple[int, int] | None = None,
        issue_line_comment: str | None = None,
    ):
        """
        :param radius_or_range: "safety buffer" to read around the issue_line
        :param issue_line_comment: add comment into one line from issue_lines,
        """
        super().__init__(filename=issue.filename, description=issue.description, issue_line=issue.issue_line)
        self.issue: Final[IssueDescriptor] = issue
        self.issue_line_comment = issue_line_comment
        with open(issue.filename, "rt") as file:
            self.all_lines: Final[list[str]] = file.read().split("\n")

        if self.issue.issue_line < 1 or self.issue.issue_line > len(self.all_lines):
            raise SystemError(
                f"Wrong issue line {self.issue.issue_line} for file `{issue.filename}`, containing"
                f" {len(self.all_lines)} lines"
            )

        if isinstance(radius_or_range, tuple):
            self._set_code_range(radius_or_range)
        elif radius_or_range is not None:
            self._set_code_radius(radius_or_range)

    def _set_code_range(self, range: tuple[int, int]):
        self.start_line = max(1, range[0])
        self.end_line = min(len(self.all_lines), range[1])
        self.safety_radius = min(self.issue.issue_line - self.start_line, self.end_line - self.issue.issue_line)
        self._set_issue_lines()

    def _set_code_radius(self, radius: int):
        self.safety_radius = abs(int(radius))
        self.start_line = max(1, self.issue.issue_line - self.safety_radius)
        self.end_line = min(len(self.all_lines), self.issue.issue_line + self.safety_radius)
        self._set_issue_lines()

    def _set_issue_lines(self):
        self.issue_lines: list[str] = copy.deepcopy(self.all_lines[self.start_line - 1 : self.end_line])
        self.proposed_lines: list[str] = copy.deepcopy(self.issue_lines)
        if self.issue_line_comment is not None:
            self.issue_lines[self.issue.issue_line - self.start_line] += self.issue_line_comment

    def try_fixing(self, query_backend: QueryBackend, diff_target: DiffTarget) -> bool:
        """
        try to find match between original issue_lines and proposed_lines
        1. We know that real change which needs to be done is in the middle of issue_lines
        2. GPT might hallucinate some starting and/or ending codes, but usually keeps provided code too
        3. We need to clear off hallucinated starting- and ending- codes, that's why we have "safety radius" around
        :return True if merge was successful
        """

        user_message_template = (
            query_backend.prompt.get("user")
            if query_backend.prompt.get("user") is not None
            else query_backend.backend.prompt.get("user")
        )

        issue_data = {
            "ISSUE_LANGUAGE": self.issue.language,
            "ISSUE_TYPE": self.issue.issue_type,
            "ISSUE_DESCRIPTION": self.issue.description,
            "ISSUE_LINES": "\n".join(self.issue_lines),
        }

        user_message = user_message_template.format(**issue_data)
        query_results: list[str] = query_backend.query(user_message, self.issue, issue_lines=self.issue_lines)
        if len(query_results) == 0:
            return False

        proposed_lines: list[str] = query_results[0].split("\n")
        merge_result = self._merge_lines(proposed_lines)

        if merge_result and self.proposed_lines != self.issue_lines:
            return diff_target.apply_diff(self)

        return False

    def _merge_lines(self, proposed_lines: list[str]) -> bool:
        if proposed_lines[0].startswith("```"):
            proposed_lines = proposed_lines[1:-1]  # remove first and last line
            if proposed_lines[-1].startswith("```"):
                # if new last line is still ``` (this happens if there was "\n" after ```)
                # remove last line once again
                proposed_lines = proposed_lines[:-1]

        # remove issue_line_comment if it was added
        issue_line_found: int | None = None
        if self.issue_line_comment is not None:
            for i in range(len(proposed_lines)):
                line: str = proposed_lines[i]
                if line.endswith(self.issue_line_comment):
                    proposed_lines[i] = line[: -len(self.issue_line_comment)]
                    issue_line_found = i
                    break

        merge_result: bool
        if issue_line_found is not None:
            merge_result = self._merge_from_issue_line(proposed_lines, issue_line_found)
        else:
            merge_result = self._merge_from_both_ends(proposed_lines)

        return merge_result

    def _merge_from_both_ends(self, proposed_lines: list[str]) -> bool:
        # merge starting code
        line_diff: list[str] = list(difflib.ndiff(self.issue_lines, proposed_lines))
        issue_index: int = 0
        prop_index: int = 0

        for i in range(len(line_diff)):  # loop until we find matching line
            if line_diff[i].startswith("+ "):
                prop_index += 1
            elif line_diff[i].startswith("- "):
                issue_index += 1
            elif line_diff[i].startswith("  "):
                proposed_lines = self.issue_lines[:issue_index] + proposed_lines[prop_index:]
                break
            else:
                prop_index += 1
                issue_index += 1

        if issue_index > self.safety_radius:
            # unsuccessful merge
            return False

        # merge ending code
        line_diff = list(difflib.ndiff(self.issue_lines, proposed_lines))
        issue_index = 0
        prop_index = 0

        for i in range(len(line_diff)):  # loop until we find matching line
            if line_diff[-i - 1].startswith("+ "):
                prop_index += 1
            elif line_diff[-i - 1].startswith("- "):
                issue_index += 1
            elif line_diff[-i - 1].startswith("  "):
                proposed_lines = proposed_lines[: -prop_index - 1] + self.issue_lines[-issue_index - 1 :]
                break
            else:
                prop_index += 1
                issue_index += 1

        if issue_index > self.safety_radius:
            # unsuccessful merge
            return False

        self.proposed_lines = proposed_lines
        return True

    def _merge_from_issue_line(self, proposed_lines: list[str], found_issue_line_index: int) -> bool:
        """
        If we can find original issue_line within the proposed_lines,
        we may use it for better merge between issue_lines and proposed_lines
        :param proposed_lines:
        :param found_issue_line_index:
        :return:
        """

        def find_first_match(list1: list[str], list2: list[str]) -> tuple[int, int]:
            index1: int = 0
            index2: int = 0
            line_diff: list[str] = list(difflib.ndiff(list1, list2))
            for i in range(len(line_diff)):  # loop until we find matching line
                if line_diff[i].startswith("+ "):
                    index2 += 1
                elif line_diff[i].startswith("- "):
                    index1 += 1
                elif line_diff[i].startswith("  "):
                    return index1, index2
                else:
                    index1 += 1
                    index2 += 1
            return index1, index2

        orig_issue_line_index: Final[int] = self.issue.issue_line - self.start_line
        issue_lines_start = self.issue_lines[:orig_issue_line_index]
        proposed_lines_start = proposed_lines[:found_issue_line_index]
        issue_lines_start.reverse()
        proposed_lines_start.reverse()

        issue_lines_end = self.issue_lines[orig_issue_line_index:]
        proposed_lines_end = proposed_lines[found_issue_line_index:]

        issue_index_above, proposed_index_above = find_first_match(issue_lines_start, proposed_lines_start)
        issue_index_below, proposed_index_below = find_first_match(issue_lines_end, proposed_lines_end)

        self.proposed_lines = (
            self.issue_lines[: orig_issue_line_index - issue_index_above]
            + proposed_lines[
                found_issue_line_index - proposed_index_above : found_issue_line_index + proposed_index_below
            ]
            + self.issue_lines[orig_issue_line_index + issue_index_below :]
        )

        return True



# Copyright: Hallux team, 2023

from __future__ import annotations

import json
from typing import Final

from ...issues.issue import IssueDescriptor
from ...proposals.proposal_engine import ProposalEngine, ProposalList
from ...proposals.python_proposal import PythonProposal
from ...proposals.simple_proposal import SimpleProposal


class SonarIssue(IssueDescriptor):
    def __init__(
        self,
        filename: str,
        issue_line: int = 0,
        description: str = "",
        text_range: dict = dict,
        issue_type: str = "warning",
        already_fixed_files: list[str] | None = None,
    ):
        super().__init__(
            language=None,
            tool="sonar",
            filename=filename,
            issue_line=issue_line,
            description=description,
            issue_type=issue_type,
        )
        if already_fixed_files is None:
            already_fixed_files = list()
        self.text_range: dict = text_range
        self.already_fixed_files: Final[list[str]] = already_fixed_files

    def list_proposals(self) -> ProposalEngine:
        start_line: int = self.text_range["startLine"]
        end_line: int = self.text_range["endLine"]
        code_range = (start_line, end_line)

        # we may only fix 1 issue per file for SonarQube
        if self.filename in self.already_fixed_files:
            return ProposalList([])

        if self.language == "python":
            line_comment: str = f" # line {str(self.issue_line)}"
            proposal_list = [
                PythonProposal(self, extract_function=True, issue_line_comment=line_comment),
                PythonProposal(self, radius_or_range=4, issue_line_comment=line_comment),
            ]
            if end_line - start_line > 4:
                proposal_list.append(PythonProposal(self, radius_or_range=code_range, issue_line_comment=line_comment))

        else:
            proposal_list = [
                SimpleProposal(self, radius_or_range=4),
                SimpleProposal(self, radius_or_range=6),
            ]
            if end_line - start_line > 4:
                proposal_list.append(SimpleProposal(self, radius_or_range=code_range))

        return ProposalList(proposal_list)

    @staticmethod
    def parseIssues(request_output: str, already_fixed_files: list[str] | None = None) -> list[SonarIssue]:
        js = json.loads(request_output)

        issues: list[SonarIssue] = []
        for js_issue in js["issues"]:
            comp_arr = js_issue["component"].split(":")
            # ToDo: do we need any extra-filter here?
            # if comp_arr != self.project:
            #     continue
            filename = comp_arr[-1]

            issue = SonarIssue(
                filename=filename,
                issue_line=js_issue["line"],
                description=js_issue["message"],
                text_range=js_issue["textRange"],
                issue_type=str(js_issue["type"]).replace("_", " "),
                already_fixed_files=already_fixed_files,
            )

            issues.append(issue)

        return issues



# Copyright: Hallux team, 2023

from __future__ import annotations

import os
from pathlib import Path
from typing import Final

from litellm import completion

from hallux.logger import logger

from ..issues.issue import IssueDescriptor
from .query_backend import QueryBackend


class LiteLLMBackend(QueryBackend):
    def __init__(
        self,
        model: str = "",
        type="litellm",
        base_path: Path = Path(),
        previous_backend: QueryBackend | None = None,
        prompt: object = None,
    ):
        assert type == "litellm"
        super().__init__(base_path, previous_backend)

        self.valid = True
        if model is None or len(model) < 2:
            logger.warning(f"Wrong model name for LiteLLM Backend: {model}")
            self.valid = False

        self.model: Final[str] = model
        self.prompt = prompt

    def query(self, request: str, issue: IssueDescriptor | None = None, issue_lines: list[str] = list) -> list[str]:
        if not self.valid:
            return []

        logger.debug("")
        logger.debug("[LiteLLM REQUEST]:")
        for line in request.split("\n"):
            logger.debug(line)

        system_message = {"content": self.prompt["system"], "role": "system"}
        user_message = {"content": request, "role": "user"}

        result = completion(model=self.model, messages=[system_message, user_message])

        answers = []

        if len(result["choices"]) > 0:
            for variant in result["choices"]:
                answers.append(variant["message"]["content"])

        logger.debug("")
        logger.debug("[LiteLLM ANSWERS]:")
        for ans in answers:
            for line in ans.split("\n"):
                logger.debug(line)

        return answers

    def validate_env(self, key: str, message: str) -> None | str:
        if os.getenv(key) is None:
            logger.warning(message.format(key=key))
            self.valid = False
        else:
            return os.getenv(key)



# Copyright: Hallux team, 2023

from __future__ import annotations

from pathlib import Path

from hallux.logger import logger
from hallux.models import PromptConfig
from ..auxilary import find_arg
from ..backends.dummy_backend import DummyBackend
from ..backends.litellm import LiteLLMBackend
from ..backends.query_backend import QueryBackend
from ..backends.rest_backend import RestBackend


class BackendFactory:
    @staticmethod
    def init_backend(argv: list[str], config: list[dict] | None, config_path: Path) -> QueryBackend:
        backends_list = config.get("backends", None)
        prompt = BackendFactory._get_prompt(config)

        default_list = [
            {"cache": {"type": "dummy", "filename": "dummy.json"}},
            {"gpt3": {"type": "litellm", "model": "gpt-3.5-turbo"}},
        ]

        backends_list: list[dict] = backends_list if backends_list is not None else default_list
        backend = None
        if not isinstance(backends_list, list) or len(backends_list) == 0:
            raise SystemError(f"BACKENDS config setting must contain non-empty list. Error in: {backends_list}")

        for name_dict in backends_list:
            name, settings = BackendFactory._validate_settings(name_dict)
            backend = BackendFactory._create_backend(settings, config_path, backend, prompt)
            if find_arg(argv, "--" + name) > 0:
                return backend  # stop early if required by CLI
        return backend

    @staticmethod
    def _validate_settings(name_dict: dict) -> tuple[str, dict[str, str]]:
        if not isinstance(name_dict, dict) or len(name_dict.items()) != 1:
            raise SystemError(f"Each BACKEND must be just one name-settings pair. Error in: {name_dict}")

        name = str(list(name_dict)[0])
        settings = name_dict[name]

        if not isinstance(settings, dict) or ("type" not in settings.keys() and not settings["model"]):
            raise SystemError(
                f"Each BACKEND setting must have at least 'type' or a 'model' setting. Error in: {settings}"
            )
        return name, settings

    @staticmethod
    def _create_backend(
        settings: dict, config_path: Path, previous_backend: QueryBackend, prompt: PromptConfig
    ) -> QueryBackend:
        type_to_class = {
            "dummy": DummyBackend,
            "rest": RestBackend,
            "litellm": LiteLLMBackend,
        }

        backend_type = settings.get("type", "").strip()

        # Set litellm as the default type if none is specified
        if not backend_type:
            backend_type = "litellm"
            logger.info("No backend type specified. Using 'litellm' as a default backend type.")

        # Add backward compatibility for openai and openai.azure types
        # TODO: remove this in the future versions
        if backend_type.lower() in ["openai", "openai.azure"]:
            backend_type = "litellm"
            settings["type"] = "litellm"
            logger.warning("'openai' and 'openai.azure' backends are deprecated. Please use 'litellm' instead.") 
        if backend_type.lower() == "openai.azure":
            settings["model"] = f"azure/{settings['model']}"
            logger.warning(f"Model name for 'openai.azure' updated to 'azure/{settings['model']}'.")

        backend_class = type_to_class.get(backend_type)
        if backend_class is None:
            raise SystemError(f"Unknown BACKEND type: {backend_type}. Supported types: {type_to_class.keys()}")
        return backend_class(**settings, base_path=config_path, previous_backend=previous_backend, prompt=prompt)

    @staticmethod
    def _get_prompt(config: dict) -> PromptConfig:
        default_system_message = """You are a senior developer who makes code reviews.
You will be given a code snippet and a description of an issue.
Fix the issue and return ONLY the fixed code, without explanations.
Keep formatting and indentation as in the original code."""

        default_user_message = """Fix "{ISSUE_LANGUAGE}" "{ISSUE_TYPE}" issue: "{ISSUE_DESCRIPTION}",
from corresponding code:\n```\n{ISSUE_LINES}\n```"""

        system_message = config.get("prompt.system", default_system_message)
        user_message = config.get("prompt.user", default_user_message)
        return {"system": system_message, "user": user_message}
